# -*- coding: utf-8 -*-
"""Subclass of `Data` to represent metadata parameters for a specific `SsspFamily`."""
from uuid import UUID

from aiida import orm
from aiida.common.lang import type_check

__all__ = ('SsspParameters',)


class SsspParameters(orm.Data):
    """Subclass of `Data` to represent metadata parameters for a specific `SsspFamily`."""

    KEY_FAMILY_UUID = 'family_uuid'

    @classmethod
    def create_from_file(cls, source, uuid):
        """Construct a new instance of metatdata parameters for an `SsspFamily` from a file.

        :param source: a filelike handle or absolute filepath.
        :param uuid: the UUID of the `SsspFamily` to which it should be coupled.
        :return: instance of `SsspParameters`.
        """
        import json

        try:
            content = source.read()
        except AttributeError:
            with open(source) as handle:
                parameters = json.load(handle)
        else:
            parameters = json.loads(content)

        return cls(parameters, uuid)

    def __init__(self, parameters, uuid, **kwargs):
        """Construct a new instance of metadata parameters for an `SsspFamily`.

        .. note:: the direction of control flows from the `SsspFamily` to the `SsspParameters`. That is to say, the
            parameters only "knows" about the `SsspFamily` through the `family_uuid` that should correspond to the
            UUID of the `SsspFamily`. However, the responsibility of validating the correspondence of the parameters
            metadata and the actual pseudos contained in the `SsspFamily` is up to the family. The `SsspParameters` will
            only guarantee that for each element that it specifies, the format of the metadata is complete and of the
            correct type. It will not validate the actual values.

        :param parameters: metadata parameters of a given `SsspFamily`. Should be a dictionary of elements, where each
            element provides a dictionary of `cutoff_wfc`, `cutoff_rho`, `filename` and `md5`.
        :param uuid: the UUID of the family to which this parameters should be coupled.
        """
        super().__init__(**kwargs)

        type_check(parameters, dict)
        type_check(uuid, (str, UUID))

        for element, values in parameters.items():
            for key, valid_types in [('filename', str), ('md5', str), ('cutoff_wfc', float), ('cutoff_rho', float)]:
                try:
                    type_check(values[key], valid_types)
                except KeyError:
                    raise ValueError('entry for element `{}` is missing the `{}` key'.format(element, key))
                except TypeError:
                    raise ValueError('`{}` for element `{}` is not of type {}'.format(key, element, valid_types))

        self.set_attribute_many(parameters)
        self.family_uuid = uuid

    def __repr__(self):
        """Represent the instance for debugging purposes."""
        return '{}<{}>'.format(self.__class__.__name__, self.pk or self.uuid)

    def __str__(self):
        """Represent the instance for human-readable purposes."""
        return self.__repr__()

    @property
    def family_uuid(self):
        """Return the UUID of the `SsspFamily` to which this parameters instance is associated.

        :return: the string UUID of the associated `SsspFamily`
        """
        return self.get_attribute(self.KEY_FAMILY_UUID)

    @family_uuid.setter
    def family_uuid(self, value):
        """Set the UUID of the `SsspFamily` to which this parameters instance is associated.

        :param value: the UUID of the associated `SsspFamily`.
        :raises: `~aiida.common.exceptions.ModificationNotAllowed`
        """
        type_check(value, (str, UUID))
        return self.set_attribute(self.KEY_FAMILY_UUID, str(value))

    @property
    def elements(self):
        """Return the set of elements defined for this instance.

        :return: set of elements
        """
        return set(self.attributes_keys()) - {self.KEY_FAMILY_UUID}

    def get_metadata(self, element=None):
        """Return the metadata for all or a specific element.

        :param element: optional element
        :raises KeyError: if the element is not defined for this instance
        """
        if element is None:
            metadata = dict(self.attributes)
            metadata.pop(self.KEY_FAMILY_UUID)
            return metadata

        try:
            return self.get_attribute(element)
        except AttributeError:
            raise KeyError('element `{}` is not defined for `{}`'.format(element, self))



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Configuration and fixtures for unit test suite."""
import json
import os
import tempfile

import pytest

pytest_plugins = ['aiida.manage.tests.pytest_fixtures']  # pylint: disable=invalid-name


@pytest.fixture
def clear_db(clear_database_before_test):
    """Alias for the `clear_database_before_test` fixture from `aiida-core`."""
    yield


@pytest.fixture
def run_cli_command():
    """Run a `click` command with the given options.

    The call will raise if the command triggered an exception or the exit code returned is non-zero
    """

    def _run_cli_command(command, options=None, raises=None):
        """Run the command and check the result.

        :param options: the list of command line options to pass to the command invocation
        :param raises: optionally an exception class that is expected to be raised
        """
        import traceback
        from click.testing import CliRunner

        runner = CliRunner()
        result = runner.invoke(command, options or [])

        if raises is not None:
            assert result.exception is not None, result.output
            assert result.exit_code != 0
        else:
            assert result.exception is None, ''.join(traceback.format_exception(*result.exc_info))
            assert result.exit_code == 0, result.output

        result.output_lines = [line.strip() for line in result.output.split('\n') if line.strip()]

        return result

    return _run_cli_command


@pytest.fixture
def uuid():
    """Return a UUID4."""
    import uuid
    return uuid.uuid4()


@pytest.fixture
def filepath_fixtures():
    """Return the absolute filepath to the directory containing the file `fixtures`."""
    return os.path.join(os.path.dirname(__file__), 'fixtures')


@pytest.fixture
def filepath_pseudos(filepath_fixtures):
    """Return the absolute filepath to the directory containing the pseudo potential files."""
    return os.path.join(filepath_fixtures, 'pseudos')


@pytest.fixture
def get_upf_data(filepath_pseudos):
    """Return `UpfData` for a given element."""

    def _get_upf_data(element='He'):
        from aiida.plugins import DataFactory
        UpfData = DataFactory('upf')
        return UpfData(os.path.join(filepath_pseudos, '{}.upf'.format(element)))

    return _get_upf_data


@pytest.fixture
def sssp_parameter_metadata():
    """Return default metadata parameters for construction of `SsspParameter`."""
    return {
        'Ar': {
            'cutoff_wfc': 10.,
            'cutoff_rho': 20.,
            'filename': 'Ar.upf',
            'md5': '62a1754735fbb79bac662092d68a8bb9'
        },
        'He': {
            'cutoff_wfc': 20.,
            'cutoff_rho': 80.,
            'filename': 'He.upf',
            'md5': '5e00510a01c97f7faa8d22f18dd6c41f'
        },
        'Ne': {
            'cutoff_wfc': 30.,
            'cutoff_rho': 240.,
            'filename': 'Ne.upf',
            'md5': 'd8afbe89a47929da4eb817a75c908077'
        }
    }


@pytest.fixture
def sssp_parameter_filepath(sssp_parameter_metadata):
    """Return a filepath that contains the parameter metadata."""
    with tempfile.NamedTemporaryFile() as filepath:
        with open(filepath.name, 'w') as handle:
            json.dump(sssp_parameter_metadata, handle)
        yield filepath.name


@pytest.fixture
def create_sssp_family(filepath_pseudos):
    """Create an `SsspFamily` from the `tests/fixtures/pseudos` directory."""

    def factory(label='SSSP/1.1/PBE/efficiency', description='SSSP v1.1 PBE efficiency'):
        from aiida_sssp.groups import SsspFamily
        return SsspFamily.create_from_folder(filepath_pseudos, label, description)

    return factory


@pytest.fixture
def create_sssp_parameters(sssp_parameter_metadata, uuid):
    """Create an `SsspParameters` from the `tests/fixtures/pseudos` directory."""

    def factory(parameters=None, uuid=uuid):
        from aiida_sssp.data import SsspParameters

        if parameters is None:
            parameters = sssp_parameter_metadata

        return SsspParameters(parameters, uuid)

    return factory


@pytest.fixture
def create_structure():
    """Return a `StructureData` instance."""

    def _create_structure(site_kind_names):
        """Return a `StructureData` instance."""
        import re
        from aiida.plugins import DataFactory

        StructureData = DataFactory('structure')

        if not isinstance(site_kind_names, (list, tuple)):
            site_kind_names = (site_kind_names,)

        structure = StructureData(cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        for kind_name in site_kind_names:
            structure.append_atom(name=kind_name, symbols=re.sub('[0-9]', '', kind_name), position=(0., 0., 0.))

        return structure

    return _create_structure



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument,pointless-statement
"""Tests for the `SsspFamily` class."""
import copy
import distutils.dir_util
import os
import shutil
import tempfile

import pytest

from aiida import orm
from aiida.common import exceptions

from aiida_sssp.data import SsspParameters
from aiida_sssp.groups import SsspFamily


def test_type_string(clear_db):
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert SsspFamily._type_string == 'sssp.family'  # pylint: disable=protected-access


def test_construct(clear_db):
    """Test the construction of `SsspFamily` works."""
    family = SsspFamily(label='SSSP').store()
    assert isinstance(family, SsspFamily)

    description = 'SSSP description'
    family = SsspFamily(label='SSSP/v1.1', description=description).store()
    assert isinstance(family, SsspFamily)
    assert family.description == description


def test_load(clear_db):
    """Test that loading of a `SsspFamily` through `load_group` works."""
    family = SsspFamily(label='SSSP').store()
    assert isinstance(family, SsspFamily)

    loaded = orm.load_group(family.pk)
    assert isinstance(family, SsspFamily)
    assert loaded.uuid == family.uuid
    assert loaded.elements == family.elements


def test_add_nodes(clear_db, get_upf_data):
    """Test the `SsspFamily.add_nodes` method."""
    upf_he = get_upf_data(element='He').store()
    upf_ne = get_upf_data(element='Ne').store()
    upf_ar = get_upf_data(element='Ar').store()
    family = SsspFamily(label='SSSP').store()

    with pytest.raises(TypeError):
        family.add_nodes(orm.Data().store())

    with pytest.raises(TypeError):
        family.add_nodes([orm.Data().store(), orm.Data().store()])

    with pytest.raises(TypeError):
        family.add_nodes([upf_ar, orm.Data().store()])

    assert family.count() == 0

    family.add_nodes(upf_he)
    assert family.count() == 1

    # Check that adding a duplicate element raises, and that no extra nodes have been added.
    with pytest.raises(ValueError):
        family.add_nodes([upf_ar, upf_he, upf_ne])
    assert family.count() == 1

    family.add_nodes([upf_ar, upf_ne])
    assert family.count() == 3


def test_elements(clear_db, get_upf_data):
    """Test the `SsspFamily.elements` property."""
    upf_he = get_upf_data(element='He').store()
    upf_ne = get_upf_data(element='Ne').store()
    upf_ar = get_upf_data(element='Ar').store()
    family = SsspFamily(label='SSSP').store()

    family.add_nodes([upf_he, upf_ne, upf_ar])
    assert family.count() == 3
    assert sorted(family.elements) == ['Ar', 'He', 'Ne']


def test_get_pseudo(clear_db, get_upf_data):
    """Test the `SsspFamily.get_pseudo` property."""
    upf_he = get_upf_data(element='He').store()
    upf_ne = get_upf_data(element='Ne').store()
    upf_ar = get_upf_data(element='Ar').store()
    family = SsspFamily(label='SSSP').store()
    family.add_nodes([upf_he, upf_ne, upf_ar])

    with pytest.raises(ValueError) as exception:
        family.get_pseudo('X')

    assert 'family `{}` does not contain pseudo for element'.format(family.label) in str(exception.value)

    element = 'He'
    upf = family.get_pseudo(element)
    assert isinstance(upf, orm.UpfData)
    assert upf.element == element


def test_validate_parameters(clear_db, create_sssp_family, create_sssp_parameters):
    """Test the `SsspFamily.validate_parameters` class method."""
    family = create_sssp_family()
    parameters = create_sssp_parameters()
    metadata = parameters.get_metadata()

    SsspFamily.validate_parameters(list(family.nodes), parameters)

    # Incorrect filename
    incorrect = copy.deepcopy(metadata['Ar'])
    incorrect['filename'] = 'wrong_file_name'
    parameters.set_attribute('Ar', incorrect)

    with pytest.raises(ValueError) as exception:
        SsspFamily.validate_parameters(list(family.nodes), parameters)

    assert 'inconsistent `filename` for element `Ar`' in str(exception.value)

    # Incorrect md5
    incorrect = copy.deepcopy(metadata['Ar'])
    incorrect['md5'] = '123abc'
    parameters.set_attribute('Ar', incorrect)

    with pytest.raises(ValueError) as exception:
        SsspFamily.validate_parameters(list(family.nodes), parameters)

    assert 'inconsistent `md5` for element `Ar`' in str(exception.value)


def test_create_from_folder(clear_db, filepath_pseudos):
    """Test the `SsspFamily.create_from_folder` class method."""
    label = 'SSSP'
    family = SsspFamily.create_from_folder(filepath_pseudos, label)

    assert isinstance(family, SsspFamily)
    assert family.is_stored
    assert family.count() == len(os.listdir(filepath_pseudos))
    assert sorted(family.elements) == sorted([filename.rstrip('.upf') for filename in os.listdir(filepath_pseudos)])

    # Cannot create another family with the same label
    with pytest.raises(ValueError):
        SsspFamily.create_from_folder(filepath_pseudos, label)

    with pytest.raises(TypeError) as exception:
        SsspFamily.create_from_folder(filepath_pseudos, label, description=1)
    assert 'Got object of type' in str(exception.value)


def test_create_from_folder_invalid(clear_db, filepath_pseudos):
    """Test the `SsspFamily.create_from_folder` class method for invalid inputs."""
    label = 'SSSP'

    with tempfile.TemporaryDirectory() as dirpath:

        # Non-existing directory should raise
        with pytest.raises(ValueError) as exception:
            SsspFamily.create_from_folder(os.path.join(dirpath, 'non-existing'), label)

        assert 'is not a directory' in str(exception.value)
        assert SsspFamily.objects.count() == 0
        assert orm.UpfData.objects.count() == 0

        distutils.dir_util.copy_tree(filepath_pseudos, dirpath)

        # Copy an existing pseudo to test that duplicate elements are not allowed
        filename = os.listdir(dirpath)[0]
        filepath = os.path.join(dirpath, filename)
        shutil.copy(filepath, os.path.join(dirpath, filename[:-4] + '2.upf'))

        with pytest.raises(ValueError) as exception:
            SsspFamily.create_from_folder(dirpath, label)

        assert 'contains pseudo potentials with duplicate elements' in str(exception.value)
        assert SsspFamily.objects.count() == 0
        assert orm.UpfData.objects.count() == 0

        # Create an empty folder in the pseudo directory, which is not allowed
        dirpath_sub = os.path.join(dirpath, 'random_sub_folder')
        os.makedirs(dirpath_sub)

        with pytest.raises(ValueError) as exception:
            SsspFamily.create_from_folder(dirpath, label)

        assert 'contains at least one entry that is not a file' in str(exception.value)
        assert SsspFamily.objects.count() == 0
        assert orm.UpfData.objects.count() == 0
        os.rmdir(dirpath_sub)

        # Create a dummy file that does not have a valid UPF format
        with open(filepath, 'w') as handle:
            handle.write('invalid pseudo format')

        with pytest.raises(ValueError) as exception:
            SsspFamily.create_from_folder(dirpath, label)

        assert 'failed to parse' in str(exception.value)
        assert SsspFamily.objects.count() == 0
        assert orm.UpfData.objects.count() == 0


def test_create_from_folder_with_parameters(clear_db, filepath_pseudos, sssp_parameter_filepath):
    """Test the `SsspFamily.create_from_folder` class method when passing a file with pseudo metadata."""
    with pytest.raises(TypeError):
        SsspFamily.create_from_folder(filepath_pseudos, 'SSSP', filepath_parameters={})

    # Test directly from filepath
    family = SsspFamily.create_from_folder(filepath_pseudos, 'SSSP/1.0', filepath_parameters=sssp_parameter_filepath)
    assert isinstance(family, SsspFamily)
    assert family.is_stored

    parameters = family.get_parameters_node()
    assert parameters.family_uuid == family.uuid

    # Test from filelike object
    with open(sssp_parameter_filepath) as handle:
        family = SsspFamily.create_from_folder(filepath_pseudos, 'SSSP/1.1', filepath_parameters=handle)
        assert isinstance(family, SsspFamily)
        assert family.is_stored

        parameters = family.get_parameters_node()
        assert parameters.family_uuid == family.uuid


def test_get_parameters_node(clear_db, create_sssp_family, create_sssp_parameters):
    """Test the `SsspFamily.get_parameters_node` method."""
    family = create_sssp_family()

    with pytest.raises(exceptions.NotExistent):
        family.get_parameters_node()

    parameters = create_sssp_parameters(uuid=family.uuid).store()

    assert isinstance(family.get_parameters_node(), SsspParameters)
    assert family.get_parameters_node().uuid == parameters.uuid


def test_parameters(clear_db, create_sssp_family, create_sssp_parameters):
    """Test the `SsspFamily.parameters` property."""
    family = create_sssp_family()

    with pytest.raises(exceptions.NotExistent):
        family.parameters

    parameters = create_sssp_parameters(uuid=family.uuid).store()

    assert isinstance(family.parameters, dict)
    assert family.parameters == parameters.attributes


def test_get_parameter(clear_db, create_sssp_family, create_sssp_parameters, sssp_parameter_metadata):
    """Test the `SsspFamily.get_parameter` method."""
    family = create_sssp_family()

    with pytest.raises(exceptions.NotExistent):
        family.get_parameter('Ar', 'cutoff')

    parameters = create_sssp_parameters(uuid=family.uuid).store()

    with pytest.raises(KeyError):
        family.get_parameter('Br', 'cutoff')

    with pytest.raises(KeyError):
        family.get_parameter('Ar', 'parameter')

    element = 'Ar'
    key_cutoff_wfc = 'cutoff_wfc'
    key_cutoff_rho = 'cutoff_rho'

    assert family.get_parameter(element, key_cutoff_wfc) == parameters.get_attribute(element)[key_cutoff_wfc]
    assert family.get_parameter(element, key_cutoff_rho) == parameters.get_attribute(element)[key_cutoff_rho]


def test_get_cutoffs(clear_db, create_sssp_family, create_sssp_parameters, create_structure):
    """Test the `SsspFamily.get_cutoffs` method."""
    family = create_sssp_family()
    parameters = create_sssp_parameters(uuid=family.uuid).store().attributes
    structure = create_structure(site_kind_names=['Ar', 'He', 'Ne'])

    with pytest.raises(ValueError):
        family.get_cutoffs(elements=None, structure=None)

    with pytest.raises(ValueError):
        family.get_cutoffs(elements='Ar', structure=structure)

    with pytest.raises(TypeError):
        family.get_cutoffs(elements=False, structure=None)

    with pytest.raises(TypeError):
        family.get_cutoffs(elements=None, structure='Ar')

    expected = parameters['Ar']
    assert family.get_cutoffs(elements='Ar') == (expected['cutoff_wfc'], expected['cutoff_rho'])
    assert family.get_cutoffs(elements=('Ar',)) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    expected = parameters['He']
    assert family.get_cutoffs(elements=('Ar', 'He')) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    expected = parameters['Ne']
    assert family.get_cutoffs(structure=structure) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    # Try structure with multiple kinds with the same element
    expected = parameters['He']
    structure = create_structure(site_kind_names=['He1', 'He2'])
    assert family.get_cutoffs(structure=structure) == (expected['cutoff_wfc'], expected['cutoff_rho'])


def test_get_pseudos(clear_db, create_sssp_family, create_sssp_parameters, create_structure):
    """Test the `SsspFamily.get_pseudos` method."""
    family = create_sssp_family()

    with pytest.raises(TypeError):
        family.get_pseudos('Ar')

    expected = {
        'Ar': family.get_pseudo('Ar'),
        'He': family.get_pseudo('He'),
        'Ne': family.get_pseudo('Ne'),
    }
    structure = create_structure(site_kind_names=['Ar', 'He', 'Ne'])
    assert family.get_pseudos(structure) == expected

    expected = {
        'Ar1': family.get_pseudo('Ar'),
        'Ar2': family.get_pseudo('Ar'),
    }
    structure = create_structure(site_kind_names=['Ar1', 'Ar2'])
    assert family.get_pseudos(structure) == expected



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Test the command line interface utilities."""
import distutils.dir_util
import enum
import os
import tarfile
import tempfile

import pytest

from aiida_sssp.cli.utils import attempt, create_family_from_archive


class ArchiveType(enum.IntEnum):
    """Simple enum that determines what type of pseudo archive will be created."""

    VALID = 0
    INVALID_ARCHIVE_FORMAT = 1
    INVALID_ARCHIVE_SUBFOLDER = 2
    INVALID_UPF_FILE = 3


@pytest.fixture
def get_pseudo_archive(filepath_pseudos, request):
    """Create a (potentially invalid) archive with pseudos."""
    archive_type, exception, message = request.param
    suffix = '.tar.gz'

    with tempfile.TemporaryDirectory() as dirpath:

        distutils.dir_util.copy_tree(filepath_pseudos, dirpath)

        if archive_type == ArchiveType.INVALID_ARCHIVE_FORMAT:
            suffix = '.txt'

        if archive_type == ArchiveType.INVALID_ARCHIVE_SUBFOLDER:
            os.makedirs(os.path.join(dirpath, 'subfolder'))

        if archive_type == ArchiveType.INVALID_UPF_FILE:
            open(os.path.join(dirpath, 'corrupt.upf'), 'a').close()

        with tempfile.NamedTemporaryFile(suffix=suffix) as filepath_archive:

            with tarfile.open(filepath_archive.name, 'w:gz') as tar:
                tar.add(dirpath, arcname='.')

            yield filepath_archive.name, exception, message


@pytest.mark.parametrize(
    'get_pseudo_archive', (
        (ArchiveType.VALID, None, None),
        (ArchiveType.INVALID_ARCHIVE_FORMAT, OSError, 'failed to unpack the archive'),
        (ArchiveType.INVALID_ARCHIVE_SUBFOLDER, OSError, 'contains at least one entry that is not a file'),
        (ArchiveType.INVALID_UPF_FILE, OSError, 'failed to parse'),
    ),
    indirect=True
)
def test_create_family_from_archive(clear_db, get_pseudo_archive, sssp_parameter_filepath):
    """Test the `create_family_from_archive` utility function."""
    from aiida_sssp.data import SsspParameters
    from aiida_sssp.groups import SsspFamily

    filepath_archive, exception, message = get_pseudo_archive

    label = 'SSSP/0.0/LDA/extreme'

    if exception is not None:
        with pytest.raises(exception) as exception:
            create_family_from_archive(label, filepath_archive)
        assert message in str(exception.value)
        return

    family = create_family_from_archive(label, filepath_archive)
    assert isinstance(family, SsspFamily)
    assert family.label == label
    assert family.count() != 0

    label = 'SSSP/1.0/LDA/extreme'
    family = create_family_from_archive(label, filepath_archive, sssp_parameter_filepath)
    assert isinstance(family, SsspFamily)
    assert family.label == label
    assert family.count() != 0
    assert isinstance(family.get_parameters_node(), SsspParameters)


def test_attempt_sucess(capsys):
    """Test the `attempt` utility function."""
    message = 'some message'

    with attempt(message):
        pass

    captured = capsys.readouterr()
    assert captured.out == 'Info: {} [OK]\n'.format(message)
    assert captured.err == ''


def test_attempt_exception(capsys):
    """Test the `attempt` utility function when exception is raised."""
    message = 'some message'
    exception = 'run-time-error'

    with pytest.raises(SystemExit):
        with attempt(message):
            raise RuntimeError(exception)

    captured = capsys.readouterr()
    assert captured.out == 'Info: {} [FAILED]\n'.format(message)
    assert captured.err == 'Critical: {}\n'.format(exception)


def test_attempt_exception_traceback(capsys):
    """Test the `attempt` utility function when exception is raised and `include_traceback=True`."""
    message = 'some message'
    exception = 'run-time-error'

    with pytest.raises(SystemExit):
        with attempt(message, include_traceback=True):
            raise RuntimeError(exception)

    captured = capsys.readouterr()
    assert captured.out == 'Info: {} [FAILED]\n'.format(message)
    assert captured.err.startswith('Critical: {}\n'.format(exception))
    assert 'Traceback' in captured.err



# -*- coding: utf-8 -*-
"""Commands to list instances of `SsspFamily`."""
import click
from aiida.cmdline.params import options as options_core
from aiida.cmdline.utils import decorators, echo

from . import options
from .root import cmd_root

PROJECTIONS_VALID = ('pk', 'uuid', 'label', 'description', 'count', 'version', 'functional', 'protocol')
PROJECTIONS_DEFAULT = ('label', 'version', 'functional', 'protocol', 'count')


def get_sssp_families_builder(version=None, functional=None, protocol=None):
    """Return a query builder that will query for SSSP families of the given configuration.

    :param version: optional version filter
    :param functional: optional functional filter
    :param protocol: optional protocol filter
    :return: `QueryBuilder` instance
    """
    from aiida.orm import QueryBuilder
    from aiida_sssp.groups import SsspFamily

    label = 'SSSP/{version}/{functional}/{protocol}'
    filters = {
        'label': {
            'like': label.format(version=version or '%', functional=functional or '%', protocol=protocol or '%')
        }
    }
    builder = QueryBuilder().append(SsspFamily, filters=filters)

    return builder


@cmd_root.command('list')
@options.VERSION(help='Filter for families with this version.')
@options.FUNCTIONAL(help='Filter for families with this functional.')
@options.PROTOCOL(help='Filter for families with this protocol.')
@options_core.PROJECT(type=click.Choice(PROJECTIONS_VALID), default=PROJECTIONS_DEFAULT)
@options_core.RAW()
@decorators.with_dbenv()
def cmd_list(version, functional, protocol, project, raw):
    """List installed configurations of the SSSP."""
    from tabulate import tabulate

    mapping_project = {
        'count': lambda family: family.count(),
        'version': lambda family: family.label.split('/')[1],
        'functional': lambda family: family.label.split('/')[2],
        'protocol': lambda family: family.label.split('/')[3],
    }

    rows = []

    for [group] in get_sssp_families_builder(version, functional, protocol).iterall():

        row = []

        for projection in project:
            try:
                projected = mapping_project[projection](group)
            except KeyError:
                projected = getattr(group, projection)
            row.append(projected)

        rows.append(row)

    if not rows:
        echo.echo_info('SSSP has not yet been installed: use `aiida-sssp install` to install it.')
        return

    if raw:
        echo.echo(tabulate(rows, disable_numparse=True, tablefmt='plain'))
    else:
        echo.echo(tabulate(rows, headers=[projection.capitalize() for projection in project], disable_numparse=True))



# -*- coding: utf-8 -*-
"""Command line interface utilities."""
from contextlib import contextmanager
from aiida.cmdline.utils import echo

__all__ = ('attempt', 'create_family_from_archive')


@contextmanager
def attempt(message, exception_types=Exception, include_traceback=False):
    """Context manager to be used to wrap statements in CLI that can throw exceptions.

    :param message: the message to print before yielding
    :param include_traceback: boolean, if True, will also print traceback if an exception is caught
    """
    import sys
    import traceback

    echo.echo_info(message, nl=False)

    try:
        yield
    except exception_types as exception:
        echo.echo_highlight(' [FAILED]', color='error', bold=True)
        message = str(exception)
        if include_traceback:
            message += '\n' + ''.join(traceback.format_exception(*sys.exc_info()))
        echo.echo_critical(message)
    else:
        echo.echo_highlight(' [OK]', color='success', bold=True)


def create_family_from_archive(label, filepath_archive, filepath_metadata=None, fmt=None):
    """Construct a new `SsspFamily` instance from a tar.gz archive.

    .. warning:: the archive should not contain any subdirectories, but just the pseudos in UPF format.

    :param label: the label for the new family
    :param filepath: absolute filepath to the .tar.gz archive containing the pseudo potentials.
    :param filepath: optional absolute filepath to the .json file containing the pseudo potentials metadata.
    :param fmt: the format of the archive, if not specified will attempt to guess based on extension of `filepath`
    :return: newly created `SsspFamily`
    :raises OSError: if the archive could not be unpacked or pseudos in it could not be parsed into a `SsspFamily`
    """
    import shutil
    import tempfile

    from aiida_sssp.groups import SsspFamily

    with tempfile.TemporaryDirectory() as dirpath:

        try:
            shutil.unpack_archive(filepath_archive, dirpath, format=fmt)
        except shutil.ReadError as exception:
            raise OSError('failed to unpack the archive `{}`: {}'.format(filepath_archive, exception))

        try:
            family = SsspFamily.create_from_folder(dirpath, label, filepath_parameters=filepath_metadata)
        except ValueError as exception:
            raise OSError('failed to parse pseudos from `{}`: {}'.format(dirpath, exception))

    return family



# -*- coding: utf-8 -*-
"""Reusable options for CLI commands."""
import click

from aiida.cmdline.params.options import OverridableOption
from aiida.cmdline.params.types import DataParamType, GroupParamType

__all__ = ('SSSP_FAMILY', 'VERSION', 'FUNCTIONAL', 'PROTOCOL')


def default_sssp_family(ctx, param, identifier):  # pylint: disable=unused-argument
    """Determine the default if no value is specified."""
    from aiida.common import exceptions
    from aiida.orm import QueryBuilder
    from aiida_sssp.groups import SsspFamily

    if identifier is not None:
        return identifier

    try:
        return QueryBuilder().append(SsspFamily).first()[0]
    except exceptions.NotExistent:
        raise click.BadParameter('failed to automatically detect an SSSP family: install it with `aiida-sssp install`.')


SSSP_FAMILY = OverridableOption(
    '-F',
    '--sssp-family',
    type=GroupParamType(sub_classes=('aiida.groups:sssp.family',)),
    required=False,
    callback=default_sssp_family,
    help='Select an SSSP family.'
)

STRUCTURE = OverridableOption(
    '-S',
    '--structure',
    type=DataParamType(sub_classes=('aiida.data:structure',)),
    help='Filter for elements of the given structure.'
)

VERSION = OverridableOption(
    '-v', '--version', type=click.STRING, required=False, help='Select the version of the SSSP configuration.'
)

FUNCTIONAL = OverridableOption(
    '-f', '--functional', type=click.STRING, required=False, help='Select the functional of the SSSP configuration.'
)

PROTOCOL = OverridableOption(
    '-p', '--protocol', type=click.STRING, required=False, help='Select the protocol of the SSSP configuration.'
)



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the command `aiida-sssp list`."""
from aiida_sssp.cli import cmd_list
from aiida_sssp.cli.list import PROJECTIONS_VALID


def test_list(clear_db, run_cli_command, create_sssp_family):
    """Test the `aiida-sssp list` command."""
    result = run_cli_command(cmd_list)
    assert 'SSSP has not yet been installed: use `aiida-sssp install` to install it.' in result.output

    family = create_sssp_family()
    result = run_cli_command(cmd_list)

    assert family.label in result.output


def test_list_raw(clear_db, run_cli_command, create_sssp_family):
    """Test the `-r/--raw` option."""
    create_sssp_family()

    for option in ['-r', '--raw']:
        result = run_cli_command(cmd_list, [option])
        assert len(result.output_lines) == 1


def test_list_project(clear_db, run_cli_command, create_sssp_family):
    """Test the `-p/--project` option."""
    family = create_sssp_family()

    # Test that all `PROJECTIONS_VALID` can actually be projected and don't except
    for option in ['-P', '--project']:
        run_cli_command(cmd_list, [option] + list(PROJECTIONS_VALID))

    result = run_cli_command(cmd_list, ['--raw', '-P', 'label'])
    assert len(result.output_lines) == 1
    assert family.label in result.output


def test_list_filter(clear_db, run_cli_command, create_sssp_family):
    """Test the filtering options `--version`, `--functional` and `--protocol`."""
    create_sssp_family(label='SSSP/1.0/PBE/efficiency')
    create_sssp_family(label='SSSP/1.1/PBEsol/precision')

    result = run_cli_command(cmd_list, ['--raw'])
    assert len(result.output_lines) == 2

    for options in [
        ('--version', '1.0'),
        ('--version', '1.1'),
        ('--functional', 'PBE'),
        ('--functional', 'PBEsol'),
        ('--protocol', 'efficiency'),
        ('--protocol', 'precision'),
    ]:
        result = run_cli_command(cmd_list, ['--raw'] + list(options))
        assert len(result.output_lines) == 1



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the command `aiida-sssp show`."""
from aiida_sssp.cli import cmd_show


def test_show(clear_db, run_cli_command, create_sssp_family):
    """Test the `aiida-sssp show` command."""
    result = run_cli_command(cmd_show, raises=SystemExit)
    assert 'Missing argument' in result.output

    family = create_sssp_family()
    result = run_cli_command(cmd_show, [family.label], raises=SystemExit)
    assert '{} does not have an associated `SsspParameters` node'.format(family) in result.output


def test_show_raw(clear_db, run_cli_command, create_sssp_family, create_sssp_parameters):
    """Test the `-r/--raw` option."""
    family = create_sssp_family()
    create_sssp_parameters(uuid=family.uuid).store()

    for option in ['-r', '--raw']:
        result = run_cli_command(cmd_show, [option, family.label])
        assert len(result.output_lines) == len(family.nodes)


def test_show_structure(clear_db, run_cli_command, create_sssp_family, create_sssp_parameters, create_structure):
    """Test the `-s/--structure` option."""
    family = create_sssp_family()
    create_sssp_parameters(uuid=family.uuid).store()
    structure = create_structure(site_kind_names=['Ar', 'He']).store()

    for option in ['-S', '--structure']:
        result = run_cli_command(cmd_show, [option, str(structure.pk), family.label])
        assert 'Ar' in result.output
        assert 'He' in result.output
        assert 'Ne' not in result.output



# -*- coding: utf-8 -*-
"""Commands to show details of particular `SsspFamily` instances."""
import click

from aiida.common import exceptions
from aiida.cmdline.params import options as options_core
from aiida.cmdline.params import types
from aiida.cmdline.utils import decorators, echo

from .root import cmd_root
from . import options


@cmd_root.command('show')
@click.argument('sssp_family', type=types.GroupParamType(sub_classes=('aiida.groups:sssp.family',)))
@options.STRUCTURE()
@options_core.RAW()
@decorators.with_dbenv()
def cmd_show(sssp_family, structure, raw):
    """Show details of a particular SSSP_FAMILY."""
    from tabulate import tabulate

    rows = []

    if structure:
        iterator = sssp_family.get_pseudos(structure).values()
    else:
        iterator = sssp_family.nodes

    try:
        sssp_family.get_parameters_node()
    except exceptions.NotExistent:
        echo.echo_critical('{} does not have an associated `SsspParameters` node'.format(sssp_family))

    for upf in iterator:
        rows.append([upf.element, upf.filename] + list(sssp_family.get_cutoffs(elements=(upf.element,))))

    headers = ['Element', 'Pseudo', 'Cutoff wfc', 'Cutoff rho']

    if raw:
        echo.echo(tabulate(sorted(rows), tablefmt='plain'))
    else:
        echo.echo(tabulate(sorted(rows), headers=headers))



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the `SsspParameters` data class."""
import json
import tempfile

import pytest

from aiida.common import exceptions
from aiida_sssp.data import SsspParameters

SSSP_PARAMETERS = {
    'Ar': {
        'cutoff_wfc': 10.,
        'cutoff_rho': 20.,
        'filename': 'Ar.upf',
        'md5': '91d02ab07c1'
    },
    'He': {
        'cutoff_wfc': 20.,
        'cutoff_rho': 80.,
        'filename': 'He.upf',
        'md5': '202f5b754e2'
    },
    'Ne': {
        'cutoff_wfc': 30.,
        'cutoff_rho': 240.,
        'filename': 'Ne.upf',
        'md5': '778fbbd8530'
    },
}


def test_construction_fail(clear_db, uuid):
    """Test the various construction arguments that should raise."""
    with pytest.raises(TypeError) as exception:
        SsspParameters([], uuid)
    assert 'Got object of type' in str(exception.value)

    with pytest.raises(TypeError) as exception:
        parameters = {'Ar': {'cutoff_rho': 2., 'filename': 'Ar.upf', 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, {'uuid'})
    assert 'Got object of type' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_rho': 2., 'filename': 'Ar.upf', 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, uuid=uuid)
    assert 'entry for element `Ar` is missing the `cutoff_wfc` key' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 1., 'filename': 'Ar.upf', 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, uuid=uuid)
    assert 'entry for element `Ar` is missing the `cutoff_rho` key' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 1., 'cutoff_rho': 2., 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, uuid=uuid)
    assert 'entry for element `Ar` is missing the `filename` key' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 1., 'cutoff_rho': 2., 'filename': 'Ar.upf'}}
        SsspParameters(parameters, uuid=uuid)
    assert 'entry for element `Ar` is missing the `md5` key' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 'str', 'cutoff_rho': 2., 'filename': 'Ar.upf', 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, uuid=uuid)
    assert '`cutoff_wfc` for element `Ar` is not of type' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 1., 'cutoff_rho': 'str', 'filename': 'Ar.upf', 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, uuid=uuid)
    assert '`cutoff_rho` for element `Ar` is not of type' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 1., 'cutoff_rho': 2., 'filename': 1., 'md5': '91d02ab07c1'}}
        SsspParameters(parameters, uuid=uuid)
    assert '`filename` for element `Ar` is not of type' in str(exception.value)

    with pytest.raises(ValueError) as exception:
        parameters = {'Ar': {'cutoff_wfc': 1., 'cutoff_rho': 2., 'filename': 'Ar.upf', 'md5': 1}}
        SsspParameters(parameters, uuid=uuid)
    assert '`md5` for element `Ar` is not of type' in str(exception.value)


def test_construction(clear_db, create_sssp_parameters):
    """Test the successful construction."""
    node = create_sssp_parameters()
    assert isinstance(node, SsspParameters)
    assert node.uuid is not None

    node.store()
    assert node.pk is not None


def test_create_from_file(clear_db, sssp_parameter_metadata, uuid):
    """Test the `SsspParameters.create_from_file` class method."""
    with tempfile.NamedTemporaryFile(mode='w') as handle:
        json.dump(sssp_parameter_metadata, handle)
        handle.flush()

        parameters = SsspParameters.create_from_file(handle.name, uuid)
        assert parameters.get_metadata() == sssp_parameter_metadata
        assert parameters.family_uuid == str(uuid)

        with open(handle.name) as source:
            parameters = SsspParameters.create_from_file(source, uuid)
            assert parameters.get_metadata() == sssp_parameter_metadata
            assert parameters.family_uuid == str(uuid)


def test_family_uuid(clear_db, create_sssp_parameters, uuid):
    """Test the `SsspParameters.family_uuid` property."""
    node = create_sssp_parameters(uuid=uuid)

    assert node.family_uuid == str(uuid)

    node.family_uuid = 'alternate'
    assert node.family_uuid == 'alternate'

    node.store()

    with pytest.raises(exceptions.ModificationNotAllowed):
        node.family_uuid = uuid


def test_elements(clear_db, create_sssp_parameters):
    """Test the `SsspParameters.elements` property."""
    node = create_sssp_parameters(parameters=SSSP_PARAMETERS)
    assert sorted(SSSP_PARAMETERS.keys()) == sorted(node.elements)


def test_get_metadata(clear_db, create_sssp_parameters):
    """Test the `SsspParameters.get_metadata` method."""
    node = create_sssp_parameters(parameters=SSSP_PARAMETERS)

    assert node.get_metadata() == SSSP_PARAMETERS

    with pytest.raises(KeyError):
        node.get_metadata('Br')

    for element, cutoffs in SSSP_PARAMETERS.items():
        assert node.get_metadata(element) == cutoffs



# -*- coding: utf-8 -*-
"""Commands to install an `SsspFamily`."""
import os

import click

from aiida.cmdline.utils import decorators, echo
from .root import cmd_root
from .utils import attempt, create_family_from_archive
from . import options

URL_BASE = 'https://legacy-archive.materialscloud.org/file/2018.0001/v4/'
URL_MAPPING = {
    ('1.0', 'PBE', 'efficiency'): 'SSSP_1.0_PBE_efficiency',
    ('1.0', 'PBE', 'precision'): 'SSSP_1.0_PBE_precision',
    ('1.1', 'PBE', 'efficiency'): 'SSSP_1.1_PBE_efficiency',
    ('1.1', 'PBE', 'precision'): 'SSSP_1.1_PBE_precision',
    ('1.1', 'PBEsol', 'efficiency'): 'SSSP_1.1_PBEsol_efficiency',
    ('1.1', 'PBEsol', 'precision'): 'SSSP_1.1_PBEsol_precision',
}


@cmd_root.command('install')
@options.VERSION(type=click.Choice(['1.0', '1.1']), default='1.1')
@options.FUNCTIONAL(type=click.Choice(['PBE', 'PBEsol']), default='PBE')
@options.PROTOCOL(type=click.Choice(['efficiency', 'precision']), default='efficiency')
@click.option('-t', '--traceback', is_flag=True, help='Include the stacktrace if an exception is encountered.')
@decorators.with_dbenv()
def cmd_install(version, functional, protocol, traceback):
    """Install a configuration of the SSSP."""
    # pylint: disable=too-many-locals
    import requests
    import tempfile

    from aiida.common import exceptions
    from aiida.common.files import md5_file
    from aiida.orm import QueryBuilder

    from aiida_sssp import __version__
    from aiida_sssp.groups import SsspFamily

    label = '{}/{}/{}/{}'.format('SSSP', version, functional, protocol)
    description = 'SSSP v{} {} {} installed with aiida-sssp v{}'.format(version, functional, protocol, __version__)

    try:
        QueryBuilder().append(SsspFamily, filters={'label': label}).limit(1).one()
    except exceptions.NotExistent:
        pass
    else:
        echo.echo_critical('SSSP {} {} {} is already installed: {}'.format(version, functional, protocol, label))

    try:
        url_base = os.path.join(URL_BASE, URL_MAPPING[(version, functional, protocol)])
    except KeyError:
        echo.echo_critical('No SSSP available for {} {} {}'.format(version, functional, protocol))

    with tempfile.TemporaryDirectory() as dirpath:

        url_archive = url_base + '.tar.gz'
        url_metadata = url_base + '.json'

        filepath_archive = os.path.join(dirpath, 'archive.tar.gz')
        filepath_metadata = os.path.join(dirpath, 'metadata.json')

        with attempt('downloading selected pseudo potentials archive... ', include_traceback=traceback):
            response = requests.get(url_archive)
            response.raise_for_status()
            with open(filepath_archive, 'wb') as handle:
                handle.write(response.content)
                handle.flush()
                description += '\nArchive pseudos md5: {}'.format(md5_file(filepath_archive))

        with attempt('downloading selected pseudo potentials metadata... ', include_traceback=traceback):
            response = requests.get(url_metadata)
            response.raise_for_status()
            with open(filepath_metadata, 'wb') as handle:
                handle.write(response.content)
                handle.flush()
                description += '\nPseudo metadata md5: {}'.format(md5_file(filepath_metadata))

        with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
            family = create_family_from_archive(label, filepath_archive, filepath_metadata)

        family.description = description
        echo.echo_success('installed `{}` containing {} pseudo potentials'.format(label, family.count()))



# -*- coding: utf-8 -*-
"""Subclass of `Group` designed to represent a family of `UpfData` nodes."""
import os

from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.orm import Group, QueryBuilder
from aiida.plugins import DataFactory

__all__ = ('SsspFamily',)

UpfData = DataFactory('upf')
SsspParameters = DataFactory('sssp.parameters')
StructureData = DataFactory('structure')


class SsspFamily(Group):
    """Group to represent a pseudo potential family.

    Each instance can only contain `UpfData` nodes and can only contain one for each element.
    """

    _node_types = (UpfData,)
    _pseudos = None
    _parameters_node = None
    _parameters = None

    def __repr__(self):
        """Represent the instance for debugging purposes."""
        return '{}<{}>'.format(self.__class__.__name__, self.pk or self.uuid)

    def __str__(self):
        """Represent the instance for human-readable purposes."""
        return '{}<{}>'.format(self.__class__.__name__, self.label)

    @classmethod
    def validate_parameters(cls, pseudos, parameters):
        """Validate the compatibility of a list of pseudos and the given metadata parameters.

        :param pseudos: list of `UpfData` nodes
        :param parameters: an instance of `SsspParameters`
        :raises ValueError: if the `SsspParameters` are not compatible with the list of pseudos
        """
        type_check(pseudos, list)
        type_check(parameters, SsspParameters)

        metadata = parameters.get_metadata()

        for pseudo in pseudos:

            type_check(pseudo, UpfData)
            element = pseudo.element

            try:
                values = metadata[element]
            except KeyError:
                raise ValueError('{} does not contain the element `{}`'.format(parameters, element))

            if values['filename'] != pseudo.filename:
                args = [parameters, 'filename', element, values['filename'], pseudo.filename]
                raise ValueError('{} inconsistent `{}` for element `{}`: {} != {}'.format(*args))

            if values['md5'] != pseudo.md5sum:
                args = [parameters, 'md5', element, values['md5'], pseudo.md5sum]
                raise ValueError('{} inconsistent `{}` for element `{}`: {} != {}'.format(*args))

    @classmethod
    def parse_pseudos_from_directory(cls, dirpath):
        """Parse the UPF files in the given directory into a list of `UpfData` nodes.

        :param dirpath: absolute path to a directory containing pseudo potentials in UPF format.
        :return: list of `UpfData` nodes
        :raises ValueError: if `dirpath` is not a directory or contains anything other than files with .UPF format
        :raises ValueError: if `dirpath` contains multiple pseudo potentials for the same element
        """
        from aiida.common.exceptions import ParsingError
        pseudos = []

        if not os.path.isdir(dirpath):
            raise ValueError('`{}` is not a directory'.format(dirpath))

        for filename in os.listdir(dirpath):
            filepath = os.path.join(dirpath, filename)

            if not os.path.isfile(filepath):
                raise ValueError('dirpath `{}` contains at least one entry that is not a file'.format(dirpath))

            try:
                pseudos.append(UpfData(filepath))
            except ParsingError as exception:
                raise ValueError('failed to parse `{}`: {}'.format(filepath, exception))

        if len(pseudos) != len(set(pseudo.element for pseudo in pseudos)):
            raise ValueError('directory `{}` contains pseudo potentials with duplicate elements'.format(dirpath))

        return pseudos

    @classmethod
    def create_from_folder(cls, dirpath, label, description=None, filepath_parameters=None):
        """Create a new `SsspFamily` from the pseudo potentials contained in a directory.

        .. note:: the directory pointed to by `dirpath` should only contain UPF files. If it contains any folders or any
            of the files cannot be parsed as valid UPF, the method will raise a `ValueError`.

        :param dirpath: absolute path to the folder containing the UPF files.
        :param label: the label to give to the `SsspFamily`, should not already exist
        :param description: optional description to give to the family.
        :param filepath_parameters: a filelike object or filepath to a file containing metadata for `SsspParameters`.
        :return: new stored instance of `SsspFamily`
        :raises ValueError: if a `SsspFamily` already exists with the given name
        """
        type_check(description, str, allow_none=True)

        try:
            cls.objects.get(label=label)
        except exceptions.NotExistent:
            family = SsspFamily(label=label)
        else:
            raise ValueError('the SsspFamily `{}` already exists'.format(label))

        pseudos = cls.parse_pseudos_from_directory(dirpath)

        if filepath_parameters is not None:
            parameters = SsspParameters.create_from_file(filepath_parameters, family.uuid)
            cls.validate_parameters(pseudos, parameters)
            parameters.store()

        if description is not None:
            family.description = description

        # Only store the `Group` and the `UpfData` nodes now, such that we don't have to worry about the clean up in
        # the case that an exception is raised during creating them.
        family.store()
        family.add_nodes([upf.store() for upf in pseudos])

        return family

    def add_nodes(self, nodes):
        """Add a node or a set of nodes to the family.

        .. note: Each family instance can only contain a single `UpfData` for each element.

        :param nodes: a single `Node` or a list of `Nodes` of type `SsspFamily._node_types`
        :raises TypeError: if nodes are not an instance or list of instance of `SsspFamily._node_types`
        :raises ValueError: if any of the elements of the nodes already exist in this family
        """
        if not isinstance(nodes, (list, tuple)):
            nodes = [nodes]

        if any([not isinstance(node, self._node_types) for node in nodes]):
            raise TypeError('only nodes of type `{}` can be added'.format(self._node_types))

        pseudos = {}

        # Check for duplicates before adding any pseudo to the internal cache
        for upf in nodes:
            if upf.element in self.elements:
                raise ValueError('element `{}` already present in this family'.format(upf.element))
            pseudos[upf.element] = upf

        self.pseudos.update(pseudos)

        super().add_nodes(nodes)

    @property
    def pseudos(self):
        """Return the dictionary of pseudo potentials of this family indexed on the element symbol.

        :return: dictionary of element symbol mapping `UpfData`
        """
        if self._pseudos is None:
            self._pseudos = {upf.element: upf for upf in self.nodes}

        return self._pseudos

    @property
    def elements(self):
        """Return the list of elements of the `UpfData` nodes contained in this family.

        :return: list of element symbols
        """
        return list(self.pseudos.keys())

    def get_pseudo(self, element):
        """Return the `UpfData` for the given element.

        :param element: the element for which to return the corresponding `UpfData` node.
        :return: `UpfData` instance if it exists
        :raises ValueError: if the family does not contain a `UpfData` for the given element
        """
        try:
            pseudo = self.pseudos[element]
        except KeyError:
            builder = QueryBuilder().append(
                SsspFamily, filters={'id': self.pk}, tag='group').append(
                self._node_types, filters={'attributes.element': element}, with_group='group')  # yapf:disable

            try:
                pseudo = builder.one()[0]
            except exceptions.MultipleObjectsError:
                raise RuntimeError('family `{}` contains multiple pseudos for `{}`'.format(self.label, element))
            except exceptions.NotExistent:
                raise ValueError('family `{}` does not contain pseudo for element `{}`'.format(self.label, element))
            else:
                self.pseudos[element] = pseudo

        return pseudo

    def get_pseudos(self, structure):
        """Return the mapping of kind names on `UpfData` for the given structure.

        :param structure: the `StructureData` for which to return the corresponding `UpfData` mapping.
        :return: dictionary of kind name mapping `UpfData`
        :raises ValueError: if the family does not contain a `UpfData` for any of the elements of the given structure.
        """
        type_check(structure, StructureData)
        return {kind.name: self.get_pseudo(kind.symbol) for kind in structure.kinds}

    def get_parameters_node(self):
        """Return the associated `SsspParameters` node if it exists.

        :return: the associated `SsspParameters` node containing information like recommended cutoffs
        :raises: `aiida.common.exceptions.NotExistent` if the family does not have associated parameters
        """
        from aiida_sssp.data import SsspParameters

        if self._parameters_node is None:
            filters = {'attributes.{}'.format(SsspParameters.KEY_FAMILY_UUID): self.uuid}
            builder = QueryBuilder().append(SsspParameters, filters=filters)
            self._parameters_node = builder.one()[0]
            self._parameters = self._parameters_node.attributes

        return self._parameters_node

    @property
    def parameters(self):
        """Return the attributes of the associated `SsspParameters` node if it exists.

        :return: a dictionary with all attributes of the associated `SsspParameters` node
        :raises: `aiida.common.exceptions.NotExistent` if the family does not have associated parameters
        """
        if self._parameters is None:
            self.get_parameters_node()

        return self._parameters

    def get_parameter(self, element, parameter):
        """Return a specific parameter for a given element.

        :param element: the element
        :param parameter: the key of the parameter
        :raises: `aiida.common.exceptions.NotExistent` if the family does not have associated parameters
        """
        try:
            element = self.parameters[element]
        except KeyError:
            raise KeyError('family `{}` does not contain the element `{}`'.format(self.label, element))

        try:
            return element[parameter]
        except KeyError:
            raise KeyError('parameter `{}` is not available for element `{}`'.format(parameter, element))

    def get_cutoffs(self, elements=None, structure=None):
        """Return the tuple of recommended cuoff and dual for either the given elements or `StructureData`.

        .. note:: at least one and only one of arguments `elements` or `structure` should be passed.

        :param elements: single or tuple of elements
        :param structure: a `StructureData` node
        :return: tuple of recommended wavefunction and density cutoff
        :raises: `aiida.common.exceptions.NotExistent` if the family does not have associated parameters
        """
        if (elements is None and structure is None) or (elements is not None and structure is not None):
            raise ValueError('at least one and only one of `elements` or `structure` should be defined')

        type_check(elements, (tuple, str), allow_none=True)
        type_check(structure, StructureData, allow_none=True)

        if structure is not None:
            symbols = structure.get_symbols_set()
        elif isinstance(elements, tuple):
            symbols = elements
        else:
            symbols = (elements,)

        cutoffs_wfc = []
        cutoffs_rho = []

        for element in symbols:
            values = self.parameters[element]
            cutoffs_wfc.append(values['cutoff_wfc'])
            cutoffs_rho.append(values['cutoff_rho'])

        return (max(cutoffs_wfc), max(cutoffs_rho))



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the command `aiida-sssp install`."""
from aiida import orm
from aiida_sssp.cli import cmd_install


def test_install(clear_db, run_cli_command):
    """Test the `aiida-sssp install` command."""
    from aiida_sssp import __version__
    from aiida_sssp.data import SsspParameters
    from aiida_sssp.groups import SsspFamily

    result = run_cli_command(cmd_install)
    assert 'installed `SSSP/' in result.output
    assert orm.QueryBuilder().append(SsspFamily).count() == 1

    family = orm.QueryBuilder().append(SsspFamily).one()[0]
    assert 'SSSP v1.1 PBE efficiency installed with aiida-sssp v{}'.format(__version__) in family.description
    assert 'Pseudo metadata md5: 0d5d6c2c840383c7c4fc3a99b5dc3001' in family.description
    assert 'Archive pseudos md5: 4803ce9fd1d84c777f87173cd4a2de33' in family.description

    parameters = family.get_parameters_node()
    assert isinstance(parameters, SsspParameters)
    assert parameters.family_uuid == family.uuid

    result = run_cli_command(cmd_install, raises=SystemExit)
    assert 'is already installed' in result.output
